# This file is part of lsst-resources.
#
# Developed for the LSST Data Management System.
# This product includes software developed by the LSST Project
# (https://www.lsst.org).
# See the COPYRIGHT file at the top-level directory of this distribution
# for details of code ownership.
#
# Use of this source code is governed by a 3-clause BSD-style
# license that can be found in the LICENSE file.

from __future__ import annotations

__all__ = ("ProxiedResourcePath",)

import contextlib
import logging
import re
from abc import ABC, abstractmethod
from collections.abc import Iterator

from ._resourcePath import ResourceHandleProtocol, ResourcePath, ResourcePathExpression
from .utils import TransactionProtocol

try:
    import fsspec
    from fsspec.spec import AbstractFileSystem
except ImportError:
    fsspec = None
    AbstractFileSystem = type


log = logging.getLogger(__name__)


class ProxiedResourcePath(ABC, ResourcePath):
    """URI that is represented internally by another type of URI for file I/O.

    For example ``abc://xyz/file.txt`` could be the public URI form but
    internally all file access is forwarded to a ``file`` URI.
    """

    _proxy: ResourcePath | None = None

    @abstractmethod
    def _set_proxy(self) -> None:
        """Calculate the internal `ResourcePath` corresponding to the public
        version.
        """
        raise NotImplementedError("Proxy must be configured")

    def _get_proxy(self) -> ResourcePath:
        """Retrieve the proxied ResourcePath."""
        proxy = self._proxy
        if proxy is None:
            raise FileNotFoundError(f"Internal error: No proxy ResourcePath available for {self}")
        return proxy

    def to_fsspec(self) -> tuple[AbstractFileSystem, str]:
        try:
            proxy = self._get_proxy()
        except FileNotFoundError:
            raise NotImplementedError(f"No proxy registered for {self}. Resource does not exist.") from None
        return proxy.to_fsspec()

    def isdir(self) -> bool:
        if self.dirLike is None:
            try:
                proxy = self._get_proxy()
            except FileNotFoundError:
                return False
            self.dirLike = proxy.isdir()
        return self.dirLike

    def exists(self) -> bool:
        try:
            proxy = self._get_proxy()
        except FileNotFoundError:
            # If there is no proxy registered then the resource can not exist.
            return False
        return proxy.exists()

    def remove(self) -> None:
        proxy = self._get_proxy()
        proxy.remove()

    def read(self, size: int = -1) -> bytes:
        proxy = self._get_proxy()
        return proxy.read(size=size)

    @contextlib.contextmanager
    def as_local(
        self, multithreaded: bool = True, tmpdir: ResourcePathExpression | None = None
    ) -> Iterator[ResourcePath]:
        proxy = self._get_proxy()
        with proxy.as_local(multithreaded=multithreaded, tmpdir=tmpdir) as loc:
            yield loc

    @contextlib.contextmanager
    def open(
        self,
        mode: str = "r",
        *,
        encoding: str | None = None,
        prefer_file_temporary: bool = False,
    ) -> Iterator[ResourceHandleProtocol]:
        proxy = self._get_proxy()
        with proxy.open(mode, encoding=encoding, prefer_file_temporary=prefer_file_temporary) as fh:
            yield fh

    def walk(
        self, file_filter: str | re.Pattern | None = None
    ) -> Iterator[list | tuple[ResourcePath, list[str], list[str]]]:
        try:
            proxy = self._get_proxy()
        except FileNotFoundError as e:
            raise ValueError(str(e)) from None
        for proxied_root, dirs, files in proxy.walk(file_filter=file_filter):
            # Need to return the directory in the original form and not the
            # proxy form.
            relative_to_self = proxied_root.path.removeprefix(proxy.path)
            root = self.replace(path=self._pathModule.join(self.path, relative_to_self))
            yield root, dirs, files

    def size(self) -> int:
        proxy = self._get_proxy()
        return proxy.size()

    def write(self, data: bytes, overwrite: bool = True) -> None:
        proxy = self._get_proxy()
        proxy.write(data, overwrite=overwrite)

    def mkdir(self) -> None:
        proxy = self._get_proxy()
        proxy.mkdir()

    def transfer_from(
        self,
        src: ResourcePath,
        transfer: str = "copy",
        overwrite: bool = False,
        transaction: TransactionProtocol | None = None,
        multithreaded: bool = True,
    ) -> None:
        proxy = self._get_proxy()
        proxy.transfer_from(
            src, transfer=transfer, overwrite=overwrite, transaction=transaction, multithreaded=multithreaded
        )



# This file is part of lsst-resources.
#
# Developed for the LSST Data Management System.
# This product includes software developed by the LSST Project
# (https://www.lsst.org).
# See the COPYRIGHT file at the top-level directory of this distribution
# for details of code ownership.
#
# Use of this source code is governed by a 3-clause BSD-style
# license that can be found in the LICENSE file.

from __future__ import annotations

__all__ = ("EupsResourcePath",)

import logging
import posixpath
import urllib.parse

from lsst.utils import getPackageDir

from ._resourcePath import ResourcePath
from .proxied import ProxiedResourcePath
from .utils import os2posix

log = logging.getLogger(__name__)


class EupsResourcePath(ProxiedResourcePath):
    """URI referring to an EUPS package.

    These URIs look like: ``eups://daf_butler/configs/file.yaml``
    where the network location is the EUPS package name.

    Internally they are proxied by either ``file`` URIs or ``resource`` URIs.
    If an ``{product}_DIR`` environment variable is found it will be used
    and internally a ``file`` URI will be created. If no environment variable
    is found an attempt will be made to convert the EUPS product name to
    a python package and a ``resource`` URI will be returned with
    ``/resources`` appended. The convention is that any package supporting
    an EUPS URI outside an EUPS environment will also have made available
    the support files as package resources.

    If it is known that the package supports package resources it is always
    better to use that URI form explicitly since it is more robust since
    not all EUPS packages can reliably be converted to python packages
    without EUPS.
    """

    quotePaths = False
    _proxy: ResourcePath | None = None
    _default_namespace: str = "lsst"

    def _set_proxy(self) -> None:
        """Calculate the internal `ResourcePath` corresponding to the public
        version.
        """
        # getPackageDir returns an absolute path.
        try:
            eups_path = getPackageDir(self.netloc)
            log.debug("Found EUPS package %s via env var", self.netloc)
        except LookupError:
            eups_path = ""
        if eups_path:
            # Must convert this path into a file URI.
            new_path = posixpath.join(os2posix(eups_path), os2posix(self.path.lstrip("/")))
            parsed = self._uri._replace(path=urllib.parse.quote(new_path), scheme="file", netloc="")
            self._proxy = ResourcePath(parsed)
            return

        # If there is no _DIR env var we need to look for python package
        # resource. There is no guaranteed way to generated a python package
        # from an EUPS product name.
        # daf_butler -> lsst.daf.butler
        # image_cutout_backend -> lsst.image_cutout_backend
        # astro_metadata_translator -> astro_metadata_translator
        product = self.netloc
        variants = (
            product,
            self._default_namespace + "." + product.replace("_", "."),
            self._default_namespace + "." + product,
            product.replace("_", "."),
        )
        for variant in variants:
            proxy = ResourcePath(f"resource://{variant}/resources", forceDirectory=True)
            # This can be slow because package imports happen but there is
            # no other way to check that we have the correct variant.
            log.debug("Trying variant %s", proxy)
            if proxy.exists():
                self._proxy = proxy
                if self.path:
                    self._proxy = self._proxy.join(self.path.lstrip("/"))
                log.debug(f"Found variant {variant}")
                return

        # Can not find an actively set up package or resources for this EUPS
        # product. Return without setting a proxy to allow standard URI
        # path manipulations to happen but defer failure until someone tries
        # to use the proxy for read.
        log.debug("Could not find any files corresponding to %s", self)
        return



# This file is part of lsst-resources.
#
# Developed for the LSST Data Management System.
# This product includes software developed by the LSST Project
# (https://www.lsst.org).
# See the COPYRIGHT file at the top-level directory of this distribution
# for details of code ownership.
#
# Use of this source code is governed by a 3-clause BSD-style
# license that can be found in the LICENSE file.

import os
import re
import sys
import unittest
import unittest.mock

from lsst.resources import ResourcePath
from lsst.resources.eups import EupsResourcePath
from lsst.resources.tests import GenericTestCase

TESTDIR = os.path.abspath(os.path.dirname(__file__))
PKG_DIR = os.path.join(TESTDIR, "packages", "eups")
PACKAGE_PATH = os.path.join(TESTDIR, "packages")


class EupsTestCase(GenericTestCase, unittest.TestCase):
    """Generic test of resource URIs."""

    scheme = "eups"
    netloc = "pkg"

    @classmethod
    def setUpClass(cls) -> None:
        # The actual value does not matter for these tests.
        os.environ["PKG_DIR"] = PKG_DIR
        super().setUpClass()

    @classmethod
    def tearDownClass(cls) -> None:
        del os.environ["PKG_DIR"]
        super().tearDownClass()

    def test_relative(self) -> None:
        # This test uses two additional netlocs which need corresponding
        # environment variables to function. The values do not matter.
        with unittest.mock.patch.dict(os.environ, {"OTHER_DIR": "x", "MY.HOST_DIR": "y"}):
            super().test_relative()


class EupsReadTestCase(unittest.TestCase):
    """Test that EUPS information can be read.

    EUPS resources can be thought of as being read only even if the
    underlying URI is a ``file`` URI.
    """

    scheme = "eups"
    netloc = "pkg"

    @classmethod
    def setUpClass(cls) -> None:
        # The actual value does not matter for these tests.
        os.environ["PKG_DIR"] = PKG_DIR
        super().setUpClass()

    @classmethod
    def tearDownClass(cls) -> None:
        del os.environ["PKG_DIR"]
        super().tearDownClass()

    def setUp(self):
        self.root = f"{self.scheme}://{self.netloc}"
        self.root_uri = ResourcePath(self.root)

    def test_read(self):
        uri = self.root_uri.join("config/test.txt")
        self.assertTrue(uri.exists(), f"Check {uri} exists")

        content = uri.read().decode()
        self.assertIn("A test config.", content)

        with uri.as_local() as local_uri:
            self.assertEqual(local_uri.scheme, "file")
            self.assertTrue(local_uri.exists())

        truncated = uri.read(size=9).decode()
        self.assertEqual(truncated, content[:9])

        # Check that directory determination can work directly without the
        # trailing slash.
        d = self.root_uri.join("config")
        self.assertTrue(d.isdir())
        self.assertTrue(d.dirLike)

        d = self.root_uri.join("config/", forceDirectory=True)
        self.assertTrue(d.exists(), f"Check directory {d} exists")
        self.assertTrue(d.isdir())

        with self.assertRaises(IsADirectoryError):
            with d.as_local() as local_uri:
                pass

        j = d.join("test.txt")
        self.assertEqual(uri, j)
        self.assertFalse(j.dirLike)
        self.assertFalse(j.isdir())
        not_there = d.join("not-there.yaml")
        self.assertFalse(not_there.exists())

        bad = ResourcePath(f"{self.scheme}://bad.module/not.yaml")
        multi = ResourcePath.mexists([uri, bad, not_there])
        self.assertTrue(multi[uri])
        self.assertFalse(multi[bad])
        self.assertFalse(multi[not_there])

        # Check that the bad URI works as expected.
        self.assertFalse(bad.exists())
        self.assertFalse(bad.isdir())
        with self.assertRaises(FileNotFoundError):
            bad.read()
        with self.assertRaises(FileNotFoundError):
            with bad.as_local():
                pass
        with self.assertRaises(FileNotFoundError):
            with bad.open("r"):
                pass

        # fsspec is always not implemented.
        with self.assertRaises(NotImplementedError):
            bad.to_fsspec()

    def test_open(self):
        uri = self.root_uri.join("config/test.txt")
        with uri.open("rb") as buffer:
            content = buffer.read()
        self.assertEqual(uri.read(), content)

        with uri.open("r") as buffer:
            content = buffer.read()
        self.assertEqual(uri.read().decode(), content)

    def test_walk(self):
        """Test that we can find file resources.

        Try to find resources in this package. Python does not care whether
        a resource is a Python file or anything else.
        """
        resource = ResourcePath(f"{self.scheme}://{self.netloc}/")
        resources = set(ResourcePath.findFileResources([resource]))

        # Do not try to list all possible options. Files can move around
        # and cache files can appear.
        subset = {
            ResourcePath(f"{self.scheme}://{self.netloc}/config/test.txt"),
            ResourcePath(f"{self.scheme}://{self.netloc}/config/test2.yaml"),
        }
        for r in subset:
            self.assertIn(r, resources)

        resources = set(
            ResourcePath.findFileResources(
                [ResourcePath(f"{self.scheme}://{self.netloc}/")], file_filter=r".*\.json"
            )
        )
        self.assertEqual(resources, set())

        # Compare regex with str.
        regex = r".*\.yaml"
        y_files_str = list(resource.walk(file_filter=regex))
        y_files_re = list(resource.walk(file_filter=re.compile(regex)))
        self.assertGreater(len(y_files_str), 1)
        self.assertEqual(y_files_str, y_files_re)

        bad_dir = ResourcePath(f"{self.scheme}://bad.module/a/dir/")
        self.assertTrue(bad_dir.isdir())
        with self.assertRaises(ValueError):
            list(bad_dir.walk())

    def test_env_var(self):
        """Test that environment variables are converted."""
        with unittest.mock.patch.dict(os.environ, {"MY_TEST_DIR": TESTDIR}):
            for env_string in ("$MY_TEST_DIR", "${MY_TEST_DIR}"):
                uri = ResourcePath(f"{env_string}/data/dir1/a.yaml")
                self.assertEqual(uri.path, "/data/dir1/a.yaml")
                self.assertEqual(uri.scheme, "eups")
                self.assertEqual(uri.netloc, "my_test")
                self.assertTrue(uri.exists())


class EupsAsResourcesReadTestCase(EupsReadTestCase):
    """Test that EUPS information can be read via resources.

    EUPS resources can be thought of as being read only even if the
    underlying URI is a ``file`` URI.
    """

    scheme = "eups"
    netloc = "pkg1"

    @classmethod
    def setUpClass(cls) -> None:
        # The actual value does not matter for these tests.
        sys.path.append(PACKAGE_PATH)
        super().setUpClass()

    @classmethod
    def tearDownClass(cls) -> None:
        sys.path.remove(PACKAGE_PATH)
        super().tearDownClass()


class EupsAsResourcesReadTestCase2(EupsAsResourcesReadTestCase):
    """Test that EUPS information can be read via resources with lsst-style
    package.

    EUPS resources can be thought of as being read only even if the
    underlying URI is a ``file`` URI.
    """

    scheme = "eups"
    netloc = "pkg2_sub"

    @classmethod
    def setUpClass(cls) -> None:
        # To avoid confusion with other lsst packages, override the default
        # prefix that is added to the EUPS name.
        cls.prefix = EupsResourcePath._default_namespace
        EupsResourcePath._default_namespace = "prefix"
        super().setUpClass()

    @classmethod
    def tearDownClass(cls) -> None:
        EupsResourcePath._default_namespace = cls.prefix
        super().tearDownClass()


if __name__ == "__main__":
    unittest.main()
